/* IRCfs - IRC FileServ for *nix.
 * Copyright (C) 2002  Nick 'Zaf' Clifford <zaf@nrc.co.nz>
 * For licensing details, refer to the LICENSE file in the source 
 * code directory.
 */
#include "runtime.h"
#include "num_hash.h"


/* TODO, make it a hash table, not a linked list! */
struct num_hash_node {
	int num;
	void *data;
	struct num_hash_node *prev, *next;
};

struct num_hash_list {
	struct num_hash_node *head,*tail;
};

struct num_hash_table {
	POOL pool;
	size_t size;
	struct num_hash_list *table;
};

struct num_hash_table *num_hash_create_n(POOL parent, size_t buckets);


static struct num_hash_node *num_hash_get_node(struct num_hash_table *ht, 
		int num);

struct num_hash_table *num_hash_create(POOL parent)
{
	return num_hash_create_n(parent,100);
}

struct num_hash_table *num_hash_create_n(POOL parent, size_t buckets)
{
	struct num_hash_table *ht;
	POOL p = pool_new(parent);
	ht = XMALLOC(p,struct num_hash_table);
	ht->pool = p;
	ht->size = buckets;
	ht->table = palloc(p,sizeof(struct num_hash_node) * buckets);
	return ht;
}
/* destroys a hash table.
 * free's any memory that *IT* allocated. eg
 * contents of nodes are *NOT* destroyed.
 * To do a total clean up, call num_hash_foreach, and free each
 * node yourself, then call num_hash_destroy
 */
int num_hash_destroy(struct num_hash_table *ht)
{
	ASSERT(ht);
	pool_destroy(ht->pool);
	return 0;
}

int num_hash_resize(struct num_hash_table *ht, size_t buckets)
{
	ASSERT(ht);
	return 0;
}

int num_hash_foreach(struct num_hash_table *ht, num_hash_callback_t callback);

int num_hash_set(struct num_hash_table *ht, int num, void *contents,
		void **old)
{
	struct num_hash_node *n;
	struct num_hash_list *hl;
	ASSERT(ht);
	n = num_hash_get_node(ht,num);
	if (n != NULL) {
		if (old)
			*old = n->data;
		n->data = contents;
		return 0;
	}
	n = XMALLOC(ht->pool, struct num_hash_node);
	hl = &ht->table[num % ht->size];
	n->prev = hl->tail;
	if (hl->tail)
		hl->tail->next = n;
	if (!hl->head) 
		hl->head = n;
	hl->tail = n;
	n->num = num;
	n->data = contents;
	return 0;


}
int num_hash_add(struct num_hash_table *ht, int num, void *contents)
{
	struct num_hash_node *n;
	struct num_hash_list *hl;
	ASSERT(ht);
	n = num_hash_get_node(ht,num);
	if (n != NULL) {
		return -1;
	}
	n = XMALLOC(ht->pool, struct num_hash_node);
	hl = &ht->table[num % ht->size];

	n->prev = hl->tail;
	if (hl->tail)
		hl->tail->next = n;
	if (!hl->head) 
		hl->head = n;
	
	hl->tail = n;
	n->next = NULL;

	n->num = num;
	n->data = contents;
	return 0;
};
void *num_hash_del(struct num_hash_table *ht, int num)
{
	struct num_hash_node *n;
	struct num_hash_list *hl;
	void *data;
	ASSERT(ht);

	n = num_hash_get_node(ht,num);
	if (!n)
		return NULL;
	hl = &ht->table[num % ht->size];
	if (n->prev)
		n->prev->next = n->next;
	else
		hl->head = n->next;
	if (n->next)
		n->next->prev = n->prev;
	else
		hl->head = n->prev;
	data = n->data;
	XFREE(n);
	return data;
}

void *num_hash_get(struct num_hash_table *ht, int num)
{
	struct num_hash_node *n;
	ASSERT(ht);
	n=num_hash_get_node(ht,num);
	if (n)
		return n->data;
	return NULL;
}

struct num_hash_node *num_hash_get_node(struct num_hash_table *ht, int num)
{
	struct num_hash_node *n;
	struct num_hash_list *hl;
	ASSERT(ht);
	hl = &ht->table[num % ht->size];
	for(n=hl->head; n != NULL; n=n->next) {
		if (n->num == num)
			return n;
	}
	return NULL;
}
